/*
 * Copyright (C)  NHR@FAU, University Erlangen-Nuremberg.
 * All rights reserved.
 * Use of this source code is governed by a MIT-style
 * license that can be found in the LICENSE file.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "parameter.h"
#include "particletracing.h"
#include "progress.h"
#include "solver.h"
#include "timing.h"

int main(int argc, char** argv)
{
    double S, E;
    Parameter params;
    Solver solver;
    ParticleTracer particletracer;
    initParameter(&params);

    if (argc != 2) {
        printf("Usage: %s <configFile>\n", argv[0]);
        exit(EXIT_SUCCESS);
    }

    readParameter(&params, argv[1]);
    printParameter(&params);
    initSolver(&solver, &params);
    printf("initsolver done\n");

    initParticleTracer(&particletracer, &solver.grid, &params);
    printParticleTracerParameters(&particletracer);

#ifndef VERBOSE
    initProgress(solver.te);
#endif

    double tau = solver.tau;
    double te  = solver.te;
    double t   = 0.0;
    int nt     = 0;

    S = getTimeStamp();
    while (t <= te) {
        if (tau > 0.0) computeTimestep(&solver);
        setBoundaryConditions(&solver);
        setSpecialBoundaryCondition(&solver);
        setObjectBoundaryCondition(&solver);
        computeFG(&solver);
        computeRHS(&solver);
        if (nt % 100 == 0) normalizePressure(&solver);
        multiGrid(&solver);

        adaptUV(&solver);

        /* Added function for particle tracing. Will inject and advance particles as per
         * timePeriod */
        trace(&particletracer, solver.u, solver.v, solver.s, t);

        t += solver.dt;
        nt++;

#ifdef VERBOSE
        printf("TIME %f , TIMESTEP %f\n", t, solver.dt);
#else
        printProgress(t);
#endif
    }
    printf("Total particles : %d\n", particletracer.totalParticles);

    E = getTimeStamp();
    stopProgress();

    freeParticles(&particletracer);

    printf("Solution took %.2fs\n", E - S);
    writeResult(&solver);
    return EXIT_SUCCESS;
}
