/*
 * Copyright (C) NHR@FAU, University Erlangen-Nuremberg.
 * All rights reserved. This file is part of nusif-solver.
 * Use of this source code is governed by a MIT style
 * license that can be found in the LICENSE file.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "vtkWriter.h"

#define U(i, j) u[(j) * (imax + 2) + (i)]
#define V(i, j) v[(j) * (imax + 2) + (i)]
#define S(i, j) s[(j) * (imax + 2) + (i)]

static int ts = 0;

void printParticles(ParticleTracer* p)
{
    for (int i = 0; i < p->totalParticles; ++i) {
        printf("Particle position X : %.2f, Y : %.2f, flag : %d\n",
            p->particlePool[i].x,
            p->particlePool[i].y,
            p->particlePool[i].flag);
    }
}
void injectParticles(ParticleTracer* p)
{
    for (int i = 0; i < p->numberOfParticles; ++i) {
        p->particlePool[p->pointer].x    = p->linSpaceLine[i].x;
        p->particlePool[p->pointer].y    = p->linSpaceLine[i].y;
        p->particlePool[p->pointer].flag = true;
        ++(p->pointer);
        ++(p->totalParticles);
    }
}

void advanceParticles(ParticleTracer* p,
    double* restrict u,
    double* restrict v,
    int* restrict s,
    double time)
{
    int imax = p->grid->imax;
    int jmax = p->grid->jmax;

    double dx = p->grid->dx;
    double dy = p->grid->dy;

    double xlength = p->grid->xlength;
    double ylength = p->grid->ylength;

    for (int i = 0; i < p->totalParticles; ++i) {
        if (p->particlePool[i].flag == true) {
            double x = p->particlePool[i].x;
            double y = p->particlePool[i].y;

            int iCoord = (int)(x / dx) + 1;
            int jCoord = (int)((y + 0.5 * dy) / dy) + 1;

            double x1 = (double)(iCoord - 1) * dx;
            double y1 = ((double)(jCoord - 1) - 0.5) * dy;
            double x2 = (double)iCoord * dx;
            double y2 = ((double)jCoord - 0.5) * dy;

            double u_n = (1.0 / (dx * dy)) *
                         ((x2 - x) * (y2 - y) * U(iCoord - 1, jCoord - 1) +
                             (x - x1) * (y2 - y) * U(iCoord, jCoord - 1) +
                             (x2 - x) * (y - y1) * U(iCoord - 1, jCoord) +
                             (x - x1) * (y - y1) * U(iCoord, jCoord));

            double new_x         = x + p->dt * u_n;
            p->particlePool[i].x = new_x;

            iCoord = (int)((x + 0.5 * dx) / dx) + 1;
            jCoord = (int)(y / dy) + 1;

            x1 = ((double)(iCoord - 1) - 0.5) * dx;
            y1 = (double)(jCoord - 1) * dy;
            x2 = ((double)iCoord - 0.5) * dx;
            y2 = (double)jCoord * dy;

            double v_n = (1.0 / (dx * dy)) *
                         ((x2 - x) * (y2 - y) * V(iCoord - 1, jCoord - 1) +
                             (x - x1) * (y2 - y) * V(iCoord, jCoord - 1) +
                             (x2 - x) * (y - y1) * V(iCoord - 1, jCoord) +
                             (x - x1) * (y - y1) * V(iCoord, jCoord));

            double new_y         = y + p->dt * v_n;
            p->particlePool[i].y = new_y;

            // printf("\tOld X : %.2f, New X : %.2f, iCoord : %d\n\tOld Y : %.2f, New Y :
            // %.2f, jCoord : %d\n\n", x, new_x, iCoord, y, new_y, jCoord);
            // printf("\tU(iCoord - 1, jCoord - 1) : %.2f, U(iCoord, jCoord - 1) : %.2f,
            // U(iCoord - 1, jCoord) : %.2f, U(iCoord, jCoord) : %.2f\n", U(iCoord - 1,
            // jCoord - 1), U(iCoord, jCoord - 1), U(iCoord - 1, jCoord), U(iCoord,
            // jCoord)); printf("\tV(iCoord - 1, jCoord - 1) : %.2f, V(iCoord, jCoord - 1)
            // : %.2f, V(iCoord - 1, jCoord) : %.2f, V(iCoord, jCoord) : %.2f\n\n",
            // V(iCoord - 1, jCoord - 1), V(iCoord, jCoord - 1), V(iCoord - 1, jCoord),
            // V(iCoord, jCoord)); printf("\t U N : %.2f, V N : %.2f\n\n", u_n, v_n);
            // printf("\t j-1 * (imax + 2) + i-1 = %d with element from U : %.2f", (jCoord
            // - 1) * (200 + 2) + (iCoord - 1), u[(jCoord - 1) * (imax + 2) + (iCoord -
            // 1)]); printf("\nimax : %d, jmax : %d\n", imax, jmax);

            if (((new_x < 0.0) || (new_x > xlength) || (new_y < 0.0) ||
                    (new_y > ylength))) {
                p->particlePool[i].flag = false;
            }
            int i_new = new_x / dx, j_new = new_y / dy;
            if (S(i_new, j_new) != NONE) {
                p->particlePool[i].flag = false;
            }
        }
    }
}

void freeParticles(ParticleTracer* p)
{
    if (p->particlePool != NULL) {
        free(p->particlePool);
        free(p->linSpaceLine);
    }
}

void writeParticles(ParticleTracer* p)
{
    VtkOptions opts = { .particletracer = p };

    char filename[50];
    // snprintf(filename, 50, "vtk_files/particles%d.vtk", ts);
    // vtkOpen(&opts, filename, ts);
    // vtkParticle(&opts, "particle");
    // vtkClose(&opts);

    FILE* fp;
    Particle* particlePool = p->particlePool;

    snprintf(filename, 50, "vis_files/particles_%d.dat", ts);
    fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("Error!\n");
        exit(EXIT_FAILURE);
    }

    for (int i = 0; i < p->totalParticles; ++i) {
        double x = particlePool[i].x;
        double y = particlePool[i].y;
        fprintf(fp, "%f %f\n", x, y);
    }
    fclose(fp);

    ++ts;
}

void initParticleTracer(ParticleTracer* p, Grid* grid, Parameter* params)
{
    p->numberOfParticles = params->numberOfParticles;
    p->startTime         = params->startTime;
    p->injectTimePeriod  = params->injectTimePeriod;
    p->writeTimePeriod   = params->writeTimePeriod;

    p->dt   = params->dt;
    p->grid = grid;

    p->x1 = params->x1;
    p->y1 = params->y1;
    p->x2 = params->x2;
    p->y2 = params->y2;

    p->lastInjectTime = params->startTime;
    p->lastUpdateTime = params->startTime;
    p->lastWriteTime  = params->startTime;

    p->pointer        = 0;
    p->totalParticles = 0;

    if (params->te > params->startTime) {
        p->estimatedNumParticles = ((params->te - params->startTime) + 2) *
                                   params->numberOfParticles;

        p->particlePool = malloc(sizeof(Particle) * p->estimatedNumParticles);
        p->linSpaceLine = malloc(sizeof(Particle) * p->numberOfParticles);

        for (int i = 0; i < p->numberOfParticles; ++i) {
            double spacing          = (double)i / (double)(p->numberOfParticles - 1);
            p->linSpaceLine[i].x    = spacing * p->x1 + (1.0 - spacing) * p->x2;
            p->linSpaceLine[i].y    = spacing * p->y1 + (1.0 - spacing) * p->y2;
            p->linSpaceLine[i].flag = true;
        }
    } else {
        p->particlePool = NULL;
        p->linSpaceLine = NULL;
    }
}

void printParticleTracerParameters(ParticleTracer* p)
{
    printf("Particle Tracing data:\n");
    printf("\tNumber of particles : %d being injected for every period of %.2f\n",
        p->numberOfParticles,
        p->injectTimePeriod);
    printf("\tstartTime : %.2f\n", p->startTime);
    printf("\t(Line along which the particles are to be injected) \n\tx1 : %.2f, y1 : "
           "%.2f, x2 : %.2f, y2 : %.2f\n",
        p->x1,
        p->y1,
        p->x2,
        p->y2);
    printf("\tPointer : %d, TotalParticles : %d\n", p->pointer, p->totalParticles);
    printf("\tdt : %.2f, dx : %.2f, dy : %.2f\n", p->dt, p->grid->dx, p->grid->dy);
}

void trace(ParticleTracer* p, double* u, double* v, int* s, double time)
{
    if (time >= p->startTime) {
        // printParticles(particletracer);
        if ((time - p->lastInjectTime) >= p->injectTimePeriod) {
            injectParticles(p);
            p->lastInjectTime = time;
        }
        if ((time - p->lastWriteTime) >= p->writeTimePeriod) {
            writeParticles(p);
            p->lastWriteTime = time;
        }
        advanceParticles(p, u, v, s, time);
        compress(p);
        p->lastUpdateTime = time;
    }
}

void compress(ParticleTracer* p)
{
    Particle* memPool = p->particlePool;
    Particle tempPool[p->totalParticles];
    int totalParticles = 0;

    for (int i = 0; i < p->totalParticles; ++i) {
        if (memPool[i].flag == 1) {
            tempPool[totalParticles].x    = memPool[i].x;
            tempPool[totalParticles].y    = memPool[i].y;
            tempPool[totalParticles].flag = memPool[i].flag;
            ++totalParticles;
        }
    }

    p->totalParticles = totalParticles;
    p->pointer        = totalParticles + 1;
    memcpy(p->particlePool, tempPool, totalParticles * sizeof(Particle));
}
